/*
    Video: https://www.youtube.com/watch?v=oCMOYS71NIU
    Based on Neil Kolban example for IDF: https://github.com/nkolban/esp32-snippets/blob/master/cpp_utils/tests/BLE%20Tests/SampleNotify.cpp
    Ported to Arduino ESP32 by Evandro Copercini

   Create a BLE server that, once we receive a connection, will send periodic notifications.
   The service advertises itself as: 6E400001-B5A3-F393-E0A9-E50E24DCCA9E
   Has a characteristic of: 6E400002-B5A3-F393-E0A9-E50E24DCCA9E - used for receiving data with "WRITE" 
   Has a characteristic of: 6E400003-B5A3-F393-E0A9-E50E24DCCA9E - used to send data with  "NOTIFY"

   The design of creating the BLE server is:
   1. Create a BLE Server
   2. Create a BLE Service
   3. Create a BLE Characteristic on the Service
   4. Create a BLE Descriptor on the characteristic
   5. Start the service.
   6. Start advertising.

   In this example rxValue is the data received (only accessible inside that function).
   And txValue is the data to be sent, in this example just a byte incremented every second. 
*/
#include <Arduino.h>
#include <BLEDevice.h>
#include <BLEServer.h>
#include <BLEUtils.h>
#include <BLE2902.h>
#include <esp32-hal-log.h>
#include <odroid_go.h>

#define CORE_DEBUG_LEVEL 5

BLEServer *pServer = NULL;
BLECharacteristic * pRTxCharacteristic;
bool deviceConnected = false;
bool oldDeviceConnected = false;
uint8_t txValue = 0;

// See the following for generating UUIDs:
// https://www.uuidgenerator.net/

#define BLE_MTU 517
#define SERVICE_UUID           "0000ffe0-0000-1000-8000-00805f9b34fb" // UART service UUID
#define CHARACTERISTIC_UUID_RTX "0000ffe1-0000-1000-8000-00805f9b34fb"

#define WHEEL_DIAMETER 27
#define MAX_SPEED 27

//uint8_t params[] = { 0x66, 0x11, 0x20, 0x09, 0x64, 0x30, 0x3C, 0x2A, 0x14, 0xD9, 0x00, 0x32, 0x00, 0x64, 0x00, 0x2D, 0x00, 0xCC, 0x10, 0x64, 0x32, 0xFC, 0x53, 0x98, 0x08, 0x64, 0x00, 0x41, 0x00, 0x00, 0x00, 0x28, 0x00, 0x00, 0x00, 0x78 };
uint8_t params[] = { 0x66, 0x11, 0x20, 0x09, 0x64, 0x30, 0x3c, 0x2a, 0x14, 0xd9, 0x00, 0x32, 0x00, 0x64, 0x00, 0x2d, 0x00, 0xcc, 0x10, 0x64, 0x32, 0xfc, 0x53, 0x98, 0x08, 0x64, 0x00, 0x41, 0x00, 0x00, 0x00, 0x28, 0x00, 0x00, 0x00, 0x78 };

byte params_request[] = { 0x66, 0x11, 0x00, 0x77 };
bool send_params = false;

uint8_t status[] = { 0x66, 0x41, 0x05, 0x05, 0x00, 0x00, 0xFF, 0x00, 0x00 };
byte status_request[] = { 0x66, 0x41, 0x02 };
bool send_status = false;

int btn_pressed = millis();
int speed = 0;
uint8_t pas_ratio = 0;

void write_status_checksum() {
  int suma = 0;
  for (int i = 0; i < sizeof(status) - 1; i++) {
    suma += status[i];
  }
  int tmp_suma = (suma / 256) * 256;
  status[sizeof(status) - 1] = suma - tmp_suma;
}

uint8_t bytes[sizeof(float)];

void gen_speed() {
  //int rpm = (int)((( speed / 60 ) * 3.141592653589793 ) / (WHEEL_DIAMETER / 39.37) * 1000 * 60 * 1000);
  int rpm = 10;
//  *(float*)(bytes) = ((WHEEL_DIAMETER / 39.37) * 3.141592653589793) * ((speed / 60) * 1000);
  Serial.println(((WHEEL_DIAMETER / 39.37) * 3.141592653589793) * ((speed / 60) * 1000));
  *(float*)(bytes) = 359.084758475;
  Serial.print("bytes: ");
  Serial.print(bytes[0], HEX);
  Serial.print(" ");
  Serial.print(bytes[1], HEX);
  Serial.print(" ");
  Serial.print(bytes[2], HEX);
  Serial.print(" ");
  Serial.print(bytes[3], HEX);
  Serial.println();
  Serial.print("### Speed: ");
  Serial.print(speed);
  Serial.print(" RPM: ");
  Serial.print(rpm);
  Serial.print(" | ");
//  status[5] = (byte) (rpm & 0xFF);
//  status[6] = (byte) ( (rpm >> 8) & 0xFF);
  status[5] = bytes[2];
  status[6] = bytes[3];
  Serial.print(status[5], HEX);
  Serial.print(" ");
  Serial.print(status[6], HEX);
  Serial.println("");
}

void write_battery() {
  GO.lcd.setCursor(10, 20);
  GO.lcd.print("Battery: ");
  GO.lcd.print(status[3]);
}

void write_fault() {
  GO.lcd.setCursor(10, 50);
  GO.lcd.print("ERROR: ");
  GO.lcd.print(status[7]);
}

void write_speed() {
  GO.lcd.setCursor(10, 80);
  GO.lcd.print("SPEED: ");
  GO.lcd.print(speed);
}

void write_pas() {
  GO.lcd.setCursor(10, 130);
  GO.lcd.print("Fahrstufe: ");
  GO.lcd.print(pas_ratio);
}

void update_display() {
  GO.lcd.clearDisplay();
  write_battery();
  write_fault();
  write_speed();
  write_pas();
}

class MyServerCallbacks: public BLEServerCallbacks {
    void onConnect(BLEServer* pServer) {
      deviceConnected = true;
    };

    void onDisconnect(BLEServer* pServer) {
      deviceConnected = false;
    }
};

class MyCallbacks: public BLECharacteristicCallbacks {
    void onWrite(BLECharacteristic *pCharacteristic) {
      std::string rxValue = pCharacteristic->getValue();
      if (rxValue.length() > 0) {
        if( (rxValue.length() == sizeof(params_request)) && (rxValue[0] == params_request[0]) && (rxValue[1] == params_request[1]) && (rxValue[2] == params_request[2]) && (rxValue[3] == params_request[3]) ) {
          send_params = true;
        }
        else if( (rxValue.length() == (sizeof(status_request) + 3)) && (rxValue[0] == status_request[0]) && (rxValue[1] == status_request[1]) && (rxValue[2] == status_request[2]) ) {
          if (pas_ratio != rxValue[3]) {
            pas_ratio = rxValue[3];
            update_display();
          }
          send_status = true;
        }
      }
    }
};

void reset_ble() {
  delay(500); // give the bluetooth stack the chance to get things ready
  pServer->startAdvertising(); // restart advertising
  oldDeviceConnected = deviceConnected;
}

void setup() {

  //esp_log_level_set("*", ESP_LOG_VERBOSE);

  GO.begin();
  GO.lcd.setTextSize(3);
  GO.lcd.clear();
  GO.Speaker.mute();
  write_battery();
  write_fault();
  write_speed();

  // Create the BLE Device
  BLEDevice::init("RAYVOLT DUMMY");
  ESP_LOGI("RayvoltDummy", "setMTU");
  BLEDevice::setMTU(BLE_MTU);
  // Create the BLE Server
  pServer = BLEDevice::createServer();
  pServer->setCallbacks(new MyServerCallbacks());

  // Create the BLE Service
  BLEService *pService = pServer->createService(SERVICE_UUID);

  // Create a BLE Characteristic
  pRTxCharacteristic = pService->createCharacteristic(
										CHARACTERISTIC_UUID_RTX,
										BLECharacteristic::PROPERTY_NOTIFY | BLECharacteristic::PROPERTY_WRITE
									);
                      
  pRTxCharacteristic->addDescriptor(new BLE2902());
  pRTxCharacteristic->setCallbacks(new MyCallbacks());
/*
  BLECharacteristic * pRxCharacteristic = pService->createCharacteristic(
											 CHARACTERISTIC_UUID_RX,
											BLECharacteristic::PROPERTY_WRITE
										);

  pRxCharacteristic->setCallbacks(new MyCallbacks());
*/
  // Start the service
  pService->start();

  // Start advertising
  pServer->getAdvertising()->start();
  reset_ble();
  ESP_LOGI("RayvoltDummy", "localMTU: %d", BLEDevice::getMTU());
}

void loop() {
  GO.update();
  if (GO.JOY_Y.isAxisPressed() == 2) {
    if (speed != MAX_SPEED) {
      speed += 1;
    }
    update_display();
  }

  if (GO.JOY_Y.isAxisPressed() == 1) {
    if (speed != 0) {
      speed -= 1;
    }
    update_display();
  }

  if ( GO.BtnA.isPressed() == 1 ) {
    if ( status[7] == 0x00 ) {
      status[7] = 0x01;
    }
    else if ( status[7] == 0x01 ) {
      status[7] = 0x02;
    }
    else if ( status[7] == 0x02 ) {
      status[7] = 0x04;
    }
    else if ( status[7] == 0x04 ) {
      status[7] = 0x08;
    }
    else if ( status[7] == 0x08 ) {
      status[7] = 0x00;
    }
    update_display();
    btn_pressed = millis();
  }

  if ( GO.BtnStart.isPressed() == 1 ) {
    if ( status[3] == 0x05 ) {
      status[3] = 0x00;
    }
    else {
      status[3] += 1;
    }
    update_display();
    btn_pressed = millis();
  }

  if (deviceConnected) {
    if (send_params) {
      ESP_LOGI("RayvoltDummy", "localMTU: %d", BLEDevice::getMTU());
      pRTxCharacteristic->setValue( { params[0], params[1], params[2], params[3], params[4], params[5], params[6], params[7], params[8], params[9], params[10], params[11] } );
      pRTxCharacteristic->notify();
      delay(10);
      pRTxCharacteristic->setValue( { params[12], params[13], params[14], params[15], params[16], params[17], params[18], params[19], params[20], params[21], params[22], params[23] } );
      pRTxCharacteristic->notify();
      delay(10);
      pRTxCharacteristic->setValue( { params[24], params[25], params[26], params[27], params[28], params[29], params[30], params[31], params[32], params[33], params[34], params[35] } );
      pRTxCharacteristic->notify();
      delay(10);
      send_params = false;
    }
    else if (send_status) {
      ESP_LOGI(millis());
      gen_speed();
      write_status_checksum();
      pRTxCharacteristic->setValue(status, sizeof(status));
      pRTxCharacteristic->notify();
      send_status = false;
    }
  delay(10); // bluetooth stack will go into congestion, if too many packets are sent
}

  // disconnecting
  if (!deviceConnected && oldDeviceConnected) {
    reset_ble();
  }
  // connecting
  if (deviceConnected && !oldDeviceConnected) {
  // do stuff here on connecting
      oldDeviceConnected = deviceConnected;
  }
}